// ASMIX Commands and Utilities
// (c) Sergei Zubkov (Cubbi) 1999
//
// pwd.s - Print Working Directory
// 
// $Id: pwd.s,v 1.1.1.1 1999/10/04 20:54:17 cubbi Exp $
//
PWD_V_MAJ = 1
PWD_V_MIN = 5

.include "config.i"
	.globl _start
_start:

.if OPTIONS
        cld
        popl    %ecx
// this was argc - 1 means we are called without arguments
        decl    %ecx
        jz      no_options

//arguments_cycle:
        popl    %edi
// this is argv
// skip argv[0]
        pushl   %ecx
        movl    $-1,%ecx
        xorl    %eax,%eax
        repne
        scasb
        popl    %ecx

// Checks for arguments
//	cmpb    $1,end_of_options
//	je      skip_options
// Check for --
//	cmpw    $0x2D2D,(%edi)
//	jne     not_dashdash
//	movb    $1,end_of_options
//	loop    arguments_cycle
not_dashdash:
// Check for -options
.if FEATURE_VERSION
// Check for -V
	cmpw    $0x562D,(%edi)
	je      show_version
.endif
// exit with error message about unrecognized option
// TBD (if +errors)
// not an option -> options no more
//	movb    $1,end_of_options

//	loop    arguments_cycle
skip_options:
no_options:
.endif


.if PWD_posix
// POSIX-compliant behaviour
// actually, it was copied from the BSD version of getcwd(), not POSIX
// and it was simplified which made it a bit slower (two syscalls per each
// directory entry in every scanned dir if readdir() is availiable)

// %ebp will hold the pointer to the discovered so far pwd
// built from right to left
	movl	$pt_buffer+1024-1,%ebp
// initialized with \n
	movl	$0x0A,(%ebp)
// %esi will hold the pointer to the ../../../../ path
	movl	$up_buffer,%esi
// initialized with ".\0"
	movw	$0x002E,up_buffer
// stat("/") to learn the root inode and dev
.if SYSCALL_unix
	pushl	$st_stat
	pushl	$root_path
	movl	$SYSCALL_STAT,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
        addl    $12,%esp
.endif
.if SYSCALL_linux
	movl	$st_stat,%ecx
	movl	$root_path,%ebx
	movl	$SYSCALL_STAT,%eax
	int	$0x80
.endif
	testl	%eax,%eax
	jnz	error_exit
// save dev and inode of the root directory
	movl	st_stat+ST_DEV,%eax
	movl	%eax,root_dev
	movl	st_stat+ST_INO,%eax
	movl	%eax,root_ino
// main cycle - over scanned directories 
main_posix_cycle:
// lstat the current "../../../../../" dir to learn its inode and dev
.if SYSCALL_unix
	pushl	$st_stat
	pushl	$up_buffer
	movl	$SYSCALL_LSTAT,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
        addl    $12,%esp
.endif
.if SYSCALL_linux
	movl	$st_stat,%ecx
	movl	$up_buffer,%ebx
	movl	$SYSCALL_LSTAT,%eax
	int	$0x80
.endif
	testl	%eax,%eax
	jnz	error_exit
// Save inode and dev of the currently scanning directory
	movl	st_stat+ST_DEV,%eax
	movl	%eax,dev
	movl	st_stat+ST_INO,%ebx
	movl	%ebx,ino
// Check for reaching root.
	cmpl	%eax,root_dev
	jne	posix_not_finished
	cmpl	%ebx,root_ino
	jne	posix_not_finished
// if root is reached - add the last '/' at the left end of pwd
	decl	%ebp
	movb	$0x2F,(%ebp)
// and printout
	jmp	posix_write
// else
posix_not_finished:
// add "..\0" to the up_buffer to make it parent
	movw	$0x2E2E,(%esi)
	incl	%esi
	incl	%esi
	movb	$0,(%esi)
// Open this parent directory O_RDONLY
.if SYSCALL_unix
	pushl	$0
	pushl	$up_buffer
	movl	$SYSCALL_OPEN,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
        addl    $12,%esp
.endif
.if SYSCALL_linux
	movl	$0,%ecx
	movl	$up_buffer,%ebx
	movl	$SYSCALL_OPEN,%eax
	int	$0x80
.endif
	testl	%eax,%eax
	jl	error_exit
	movl	%eax,%edi
.if SYSCALL_unix
	pushl	$st_stat
	pushl	%eax
	movl	$SYSCALL_FSTAT,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
        addl    $12,%esp
.endif
.if SYSCALL_linux
// stat this parent directory
	movl	$st_stat,%ecx
	movl	%eax,%ebx
	movl	$SYSCALL_FSTAT,%eax
	int	$0x80
.endif
	testl	%eax,%eax
	jnz	error_exit
// If all was ok, add trailing slash for next directory.
	movb	$0x2F,(%esi)
	incl	%esi
// cycle over contents of the parent directory - we are looking for the entry
// which has the same inode and device as the saved inode and dev of the current
	xorl	%ebx,%ebx
readdir_cycle:
// readdir() - shorter to use than getdents
.if SYSCALL_READDIR
.if SYSCALL_unix
	pushl	$1
	pushl	$dirent
	pushl	%edi
	movl	$SYSCALL_READDIR,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
        addl    $16,%esp
.endif
.if SYSCALL_linux
	movl	$1,%edx
	movl	$dirent,%ecx
	movl	%edi,%ebx
	movl	$SYSCALL_READDIR,%eax
	int	$0x80
.endif
 # ebx will hold the dirent we are currenlty working on
	movl	$dirent,%ebx
	decl	%eax
	jne	notfound

.else

// what to do if no syscall_readdir (but have syscall_getdents or getdirentries)
	testl	%ebx,%ebx
	jz	time_to_getdent
	xorl	%eax,%eax
	movw	DIRENT_LEN(%ebx),%ax
	addl	%eax,%ebx
// check if ebx will be beyond the boundaries after adding LEN
	xorl	%eax,%eax
	movw	DIRENT_LEN(%ebx),%ax
	cmpl	dirent_filled,%eax
	jge	time_to_getdent
	movw	DIRENT_LEN(%ebx),%ax
	testl	%eax,%eax
	jnz	skip_getdent
time_to_getdent:
.if SYSCALL_unix
.if SYSCALL_GETDENTS
	pushl	$SIZE_DIRENT
	pushl	$dirent
	pushl	%edi
	movl	$SYSCALL_GETDENTS,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
        addl    $16,%esp
.else
	pushl	$basep
	pushl	$SIZE_DIRENT
	pushl	$dirent
	pushl	%edi
	movl	$SYSCALL_GETDIRENTRIES,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
	addl	$20,%esp
.endif
.endif
.if SYSCALL_linux
	movl	$SIZE_DIRENT,%edx
	movl	$dirent,%ecx
	movl	%edi,%ebx
	movl	$SYSCALL_GETDENTS,%eax
	int	$0x80
.endif
	movl	%eax,dirent_filled
	testl	%eax,%eax
	je	notfound
	movl	$dirent,%ebx
skip_getdent:
.endif
// copy the name of the entry on the right end of ../../../../ line
// not moving the pointer in %esi
	xorl	%ecx,%ecx
	pushl	%esi
	pushl	%edi
	movl	%esi,%edi
	movl	%ebx,%esi
	addl	$DIRENT_NAME,%esi
	movw	DIRENT_LEN(%ebx),%cx
	incl	%ecx
	rep
	movsb
	popl	%edi
	popl	%esi
// lstat this entry to get its inode and dev
	pushl	%ebx
.if SYSCALL_unix
	pushl	$st_stat
	pushl	$up_buffer
	movl	$SYSCALL_LSTAT,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
        addl    $12,%esp
.endif
.if SYSCALL_linux
	movl	$st_stat,%ecx
	movl	$up_buffer,%ebx
	movl	$SYSCALL_LSTAT,%eax
	int	$0x80
.endif
	popl	%ebx
	testl	%eax,%eax
	jnz	readdir_cycle
// if (s.st_dev == dev && s.st_ino == ino)
// if they are not the same as the saved current ino and dev - continue
// readdirs
	movl	dev,%eax
	cmpl	%eax,st_stat+ST_DEV
	jne	readdir_cycle
	movl	ino,%eax
	cmpl	%eax,st_stat+ST_INO
	jne	readdir_cycle
// return up_buffer to the state of ../../../../../ equal to the parent dir
	movl $0,(%esi)

// put a / into the creating pwd line except if it was the very first dir
	cmpb	$1,first
	je	not_first
	decl	%ebp
	movb	$0x2F,(%ebp)
not_first:
	movb	$0,first
// move the %ebp pointer namlen bytes left
	xorl	%edx,%edx
.if SYSCALL_READDIR
	movw	DIRENT_LEN(%ebx),%dx
.else
	pushl	%esi
	xorl	%ecx,%ecx
	movl	%ebx,%esi
	addl	$DIRENT_NAME,%esi
seek_zero:
	incl	%ecx
	movb	(%esi,%ecx,1),%al
	testb	%al,%al
	jnz	seek_zero
	movl	%ecx,%edx
	popl	%esi
.endif
	subl	%edx,%ebp
// copy the name of the current direntry into the pwd, not moving the pointer
	xorl	%ecx,%ecx
	pushl	%esi
	pushl	%edi
	movl	%ebp,%edi
	movl	%ebx,%esi
	addl	$DIRENT_NAME,%esi
.if SYSCALL_READDIR
	movw	DIRENT_LEN(%ebx),%cx
.else
	movl	%edx,%ecx
.endif
	rep
	movsb
	popl	%edi
	popl	%esi
// close the directory
.if SYSCALL_unix
	pushl	%edi
	movl	$SYSCALL_CLOSE,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
        addl    $8,%esp
.endif
.if SYSCALL_linux
	movl	%edi,%ebx
	movl	$SYSCALL_CLOSE,%eax
	int	$0x80
.endif
// continue untill we get to root
	jmp	main_posix_cycle
notfound:
	jmp error_exit
posix_write:
// print out the output line  - write(1,buf,count)
	movl	$pt_buffer+1024,%edx
	subl	%ebp,%edx
	movl	%ebp,%ecx
	xorl	%ebx,%ebx
	incl	%ebx
.if SYSCALL_unix
	pushl	%edx
	pushl	%ecx
	pushl	%ebx
	movl	$SYSCALL_WRITE,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
        addl    $16,%esp
.endif
.if SYSCALL_linux
	movl	$SYSCALL_WRITE,%eax
	int	$0x80
.endif
.endif

// print out the PWD environment variable - only included for
// completeness, DON'T USE!
.if PWD_env
	cld
	popl	%ebx
	movl	4(%esp,%ebx,4),%edi
// *envp now in %edi
	xorl	%eax,%eax
env_scan_cycle:
	cmpl	$0,(%edi)
	jz	error_exit
	movl	$env_name,%esi
	movl	$env_name_l,%ecx
	repe
	cmpsb
	jcxz    found_variable
	xorl	%ecx,%ecx
	decl	%ecx
	repne
	scasb
	jmp     env_scan_cycle

// %edi - address of the zero-terminated PWD value
found_variable:
	xorl	%edx,%edx
	decl	%edx
loop_zero_scan:
	incl	%edx
	cmpb	$0,(%edi,%edx,1)
	jne	loop_zero_scan
	movb	$0x0A,(%edi,%edx,1)
	incl	%edx
.if SYSCALL_unix
	pushl	%edx
	pushl	%edi
	pushl	$1
	movl	$SYSCALL_WRITE,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
        addl    $16,%esp
.endif
.if SYSCALL_linux
	movl	%edi,%ecx
	movl	$1,%ebx
	movl	$SYSCALL_WRITE,%eax
	int	$0x80
.endif
.endif

.if PWD_proc
// sys_readlink("/proc/self/cwd")
	movl	$255,%edx
	movl	$linux_buffer,%ecx
	movl	$linux_pwd,%ebx
	movl	$SYSCALL_READLINK,%eax
	int	$0x80
.endif

.if PWD_sys
// sys_getcwd()
.if SYSCALL_unix
	pushl	$MAX_PATH
	pushl	$linux_buffer
	movl	$SYSCALL_GETCWD,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
        addl    $12,%esp
.endif
.if SYSCALL_linux
	movl	$MAX_PATH,%ecx
	movl	$linux_buffer,%ebx
	movl	$SYSCALL_GETCWD,%eax
	int	$0x80
.endif
	testl	%eax,%eax
	js	error_exit
	decl	%eax
// really for FreeBSD __getcwd() syscall
.if SYSCALL_unix
	xorl	%eax,%eax
	xorl	%ecx,%ecx
	decl	%ecx
	movl	$linux_buffer,%edi
	repne
	scasb
	subl	$linux_buffer,%edi
	movl	%edi,%eax
	incl	%eax
.endif
.endif

// add CR at the end of the output line
.if PWD_proc | PWD_sys
	movb	$0x0A,linux_buffer(%eax)
	incl	%eax

// print out the output line  - write(1,buf,count)
.if SYSCALL_unix
	pushl	%eax
	pushl	$linux_buffer
	pushl	$1
	movl	$SYSCALL_WRITE,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
        addl    $16,%esp
.endif
.if SYSCALL_linux
	movl	%eax,%edx
	xorl	%ebx,%ebx
	movl	$linux_buffer,%ecx
	incl	%ebx
	movl	$SYSCALL_WRITE,%eax
	int	$0x80
.endif
.endif

// exit the program - exit(0)
exit:
	xorl	%ebx,%ebx
.if SYSCALL_unix
	pushl	%ebx
	movl	$SYSCALL_EXIT,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
.endif
.if SYSCALL_linux
	movl	$SYSCALL_EXIT,%eax
	int	$0x80
.endif

.if ~(-PWD_proc)
exit_errno:
exit_ENOENT:
error_exit:
.if SYSCALL_unix
	pushl	$1
	movl	$SYSCALL_EXIT,%eax
	pushl	%eax
        .byte   0x9A
        .long   0
        .word   7
.endif
.if SYSCALL_linux
	xorl	%ebx,%ebx
	incl	%ebx
	movl	$SYSCALL_EXIT,%eax
	int	$0x80
.endif
.endif

	.data
.if PWD_posix
pwd_path:
	.ascii	".\000"
root_path:
	.ascii	"/\000"
first:
	.byte	1
.endif

// the special file on /proc filesystem that is a soft link to current dir
.if PWD_proc
linux_pwd:
	.ascii	"/proc/self/cwd\000"
.endif

// the environment variable name for the env version
.if PWD_env
env_name:
	.ascii	"PWD="
env_name_l = .-env_name
.endif

.if SYSCALL_GETDIRENTRIES
basep:
        .long   0
.endif

	.bss

// buffer for stat()
.if PWD_posix
	.lcomm	st_stat,SIZE_STAT
.endif

// buffer to fill with the sys_getcwd or with readlink
.if PWD_proc | PWD_sys
	.lcomm	linux_buffer,MAX_PATH
.endif

// buffer for the pwd output line and for ../../../../
.if PWD_posix
	.lcomm	pt_buffer,1024
	.lcomm	up_buffer,1024
.endif

// various variables
.if PWD_posix
	.lcomm	root_dev,4
	.lcomm	dev,4
	.lcomm	root_ino,4
	.lcomm	ino,4
	.lcomm	bpt,4
// buffer for dirent
	.lcomm	dirent,SIZE_DIRENT
.if SYSCALL_GETDENTS | SYSCALL_GETDIRENTRIES
	.lcomm	dirent_filled,4
.endif
.endif


	.text
.if FEATURE_VERSION
show_version:
.if SYSCALL_unix
	pushl   $version_l
	pushl   $version
	pushl   $1
	movl    $SYSCALL_WRITE,%eax
	pushl   %eax
	.byte   0x9A
	.long   0
	.word   7
	addl    $16,%esp
.endif
.if SYSCALL_linux
	movl    $version_l,%edx
	movl    $version,%ecx
	movl    $1,%ebx
	movl    $SYSCALL_WRITE,%eax
	int     $0x80
.endif
	jmp     exit
.endif

	.data
.if OPTIONS
end_of_options:
        .byte   0
.endif
.if FEATURE_VERSION
version:
        .ascii  "pwd.s "
        .byte   '0'+PWD_V_MAJ
        .ascii  "."
        .byte   '0'+PWD_V_MIN
.if PWD_sys
	.ascii	"/syscall"
.endif
.if PWD_env
	.ascii	"/environ"
.endif
.if PWD_posix
	.ascii	"/posix"
.endif
.if PWD_proc
	.ascii	"/proc"
.endif
        .ascii  "+version"
// +other things here
        .ascii  " ASMIX CU "
        .byte   '0'+ASMIX_VER_MAJ
        .ascii  "."
        .byte   '0'+ASMIX_VER_MIN
        .ascii  "."
        .byte   '0'+ASMIX_VER_REV
        .ascii  "\012"
version_l = .-version
.endif



